#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# File: image_processing.py
# Project: Optimized Registration through Conditional Adversarial networks (ORCA)
# Author: Zacharias Chalampalakis | Lalith Kumar Shiyam Sundar | Sebastian Gutschmayer
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 19.07.2023
# Version: 0.1.0
#
# Description:
# This module handles image processing for orca.
#
# Usage:
# The functions in this module can be imported and used in other modules within orca to perform image conversion.
# ----------------------------------------------------------------------------------------------------------------------

import SimpleITK as sitk
import contextlib
import time
import glob
from orcaz import constants
from orcaz.constants import GREEDY_PATH
from orcaz.file_utilities import *
from mpire import WorkerPool
import multiprocessing
from rich.progress import Progress
from concurrent.futures import ThreadPoolExecutor
from moosez import moose
import os
import nibabel as nib
import numpy as np
import pathlib
import subprocess
import logging
import sys
import re
from orcaz import resources
from rich.progress import track
import nibabel
from scipy.ndimage import binary_dilation
from nifti2dicom.converter import save_dicom_from_nifti_image

DEBUG_MODE=False


def process_and_moose_ct_files(ct_dir: str, mask_dir: str, moose_model: str, accelerator: str):
    # Get the ct_files in the ct_dir that are resampled to the PET space
    ct_files = get_files(ct_dir, 'resampled*.nii*')

    with Progress() as progress_bar:
        task = progress_bar.add_task("[cyan] MOOSE-ing CT files...", total=len(ct_files))
        
        for ct_file in ct_files:
            base_name = os.path.basename(ct_file).split('.')[0]
            
            ct_file_dir = os.path.join(ct_dir, base_name)
            create_directory(ct_file_dir)
            move_file(ct_file, os.path.join(ct_file_dir, os.path.basename(ct_file)))
            
            mask_file_dir = os.path.join(mask_dir, base_name)
            create_directory(mask_file_dir)
            
            progress_bar.update(task, advance=0, description=f"[cyan] Running MOOSE on {base_name}...")

            ct_file = glob.glob(os.path.join(ct_file_dir,"*.nii*"))[0]
            
            with open(os.devnull, 'w') as devnull:
                with contextlib.redirect_stdout(devnull), contextlib.redirect_stderr(devnull):
                    moose(model_names="clin_ct_body",
                          input_data=ct_file,
                          accelerator=accelerator,
                          output_dir=mask_file_dir)

            progress_bar.update(task, advance=0, description=f"[cyan] Completed MOOSE on {base_name}, cleaning up...")

            move_files_to_directory(ct_file_dir, ct_dir)
            move_files_to_directory(mask_file_dir, mask_dir)

            remove_directory(ct_file_dir)
            remove_directory(mask_file_dir)

            progress_bar.update(task, advance=1, description=f"[cyan] Finished processing {base_name}")

            time.sleep(1)  # delay for 1 second

def change_mask_labels(mask_file: str, label_map: dict, excluded_labels: list):
    # Load the image
    img = nib.load(mask_file)

    # Get the image data (returns a numpy array)
    data = img.get_fdata()

    # Prepare labels for modification
    excluded_indices = [idx for idx, lbl in label_map.items() if lbl in excluded_labels]
    other_indices = [idx for idx, lbl in label_map.items() if lbl not in excluded_labels]

    # Set the labels
    data[np.isin(data, excluded_indices)] = 0
    data[np.isin(data, other_indices)] = 1

    # Save the modified image
    new_img = nib.Nifti1Image(data, img.affine, img.header)
    nib.save(new_img, mask_file)

def reslice_identity(reference_image: sitk.Image, moving_image: sitk.Image, 
                     output_image_path: str = None, is_label_image: bool = False, median=-1024) -> sitk.Image:
    """
    Reslice an image to the same space as another image
    :param reference_image: The reference image
    :param moving_image: The image to reslice to the reference image
    :param output_image_path: Path to the resliced image
    :param is_label_image: Determines if the image is a label image. Default is False
    """
    resampler = sitk.ResampleImageFilter()
    resampler.SetReferenceImage(reference_image)
    resampler.SetDefaultPixelValue(int(median))

    if is_label_image:
        resampler.SetInterpolator(sitk.sitkNearestNeighbor)
    else:
        resampler.SetInterpolator(sitk.sitkBSpline)

    resampled_image = resampler.Execute(moving_image)
    resampled_image = sitk.Cast(resampled_image, sitk.sitkInt32)
    if output_image_path is not None:
        sitk.WriteImage(resampled_image, output_image_path)
    return resampled_image


def prepare_reslice_tasks(subject_directory: str):
    tasks = []
    ct_file = glob.glob(os.path.join(subject_directory, '*CT*.nii*'))
    pt_file = glob.glob(os.path.join(subject_directory, '*PT*.nii*'))
    resliced_ct_file = os.path.join(subject_directory, constants.RESAMPLED_PREFIX + '_' +
                                    os.path.basename(ct_file[0]))

    tasks.append((
        sitk.ReadImage(pt_file[0]),
        sitk.ReadImage(ct_file[0]),
        resliced_ct_file,
        False
    ))
    return tasks


def copy_and_rename_file(src, dst, subdir):
    copy_file(src, dst)
    new_file = os.path.join(dst, os.path.basename(subdir) + '_' + os.path.basename(src))
    os.rename(os.path.join(dst, os.path.basename(src)), new_file)


def preprocess(subject_dir: str, dicom_files_flag: bool, num_workers: int = None, coreg_flag: bool = True):
    """
    Preprocesses the images in the subject directory
    :param puma_compliant_subjects: The puma compliant subjects
    :param num_workers: The number of worker processes for parallel processing
    """
    if num_workers is None:
        num_workers = multiprocessing.cpu_count()

    pt_file = glob.glob(os.path.join(subject_dir, '*PT*.nii*'))
    if coreg_flag:
        ct_file = glob.glob(os.path.join(subject_dir, '*CT*.nii*'))
        resliced_ct_file = os.path.join(subject_dir, constants.RESAMPLED_PREFIX + '_' +
                                        os.path.basename(ct_file[0]))
        reslice_identity(sitk.ReadImage(pt_file[0]), sitk.ReadImage(ct_file[0]), resliced_ct_file, is_label_image=False)

    # set orca working directory
 
    orca_working_dir = os.path.join(os.path.abspath(subject_dir), constants.ORCA_WORKING_FOLDER)
    create_directory(orca_working_dir)

    # create CT and PET folders
    if coreg_flag:
        ct_dir = os.path.join(orca_working_dir, constants.MODALITIES[1])
    else:
        ct_dir=[]
    pt_dir = os.path.join(orca_working_dir, constants.MODALITIES[0])
    mask_dir = os.path.join(orca_working_dir, constants.MASK_FOLDER)
    if coreg_flag: create_directory(ct_dir)
    create_directory(pt_dir)


    # Move or copy and rename files
    ct_files = glob.glob(os.path.join(subject_dir, '*CT*.nii*'))
    if coreg_flag: 
        for ct_file in ct_files: move_file(ct_file, ct_dir) if dicom_files_flag else copy_file(ct_file, ct_dir)
    pt_files = glob.glob(os.path.join(subject_dir, '*PT*.nii*'))
    for pt_file in pt_files: move_file(pt_file, pt_dir) if dicom_files_flag else copy_file(pt_file, pt_dir)

    if coreg_flag:
        # Run moosez to get the masks
        process_and_moose_ct_files(ct_dir, mask_dir, constants.MOOSE_MODEL, constants.ACCELERATOR)

        # remove the prefix from the mask files
        for mask_file in glob.glob(os.path.join(mask_dir, constants.MOOSE_PREFIX + '*')):
            new_mask_file = re.sub(rf'{constants.MOOSE_PREFIX}', '', mask_file)
            os.rename(mask_file, new_mask_file)
            # for ORCA the whole body contour is needed, so no masks are removed. 
            #change_mask_labels(new_mask_file, constants.MOOSE_LABEL_INDEX, ["Arms"])

        # process the mask files for ORCA
        process_mask(mask_dir)

    return orca_working_dir, ct_dir, pt_dir, mask_dir

def postprocess(orca_working_dir: str, ct_dir: int = None, mask_dir: int = None):
    ct_file = glob.glob(os.path.join(ct_dir, '*_CT.nii*'))[0]
    resampled_moving_img = glob.glob(os.path.join(orca_working_dir, constants.ALIGNED_CT_FOLDER + "/ORCA*.nii*"))[0]
    _ = reslice_identity(sitk.ReadImage(ct_file), sitk.ReadImage(resampled_moving_img),
                        os.path.join(orca_working_dir, constants.ALIGNED_CT_FOLDER + "/ORCA_aligned_CT.nii.gz"),
                        is_label_image=False)
    
    resampled_moving_mask = glob.glob(os.path.join(mask_dir, "ORCA*.nii*"))[0]
    _ = reslice_identity(sitk.ReadImage(ct_file), sitk.ReadImage(resampled_moving_mask),
                        os.path.join(os.path.abspath(mask_dir), "ORCA_aligned_mask.nii.gz"),
                        is_label_image=True, median=0)
    
    # Clean up the edges of the CT images 
    for ct_file in glob.glob(os.path.join(orca_working_dir, constants.ALIGNED_CT_FOLDER + "/*.nii*")): 
        #ToDo Make this search better, filenaming is not consistent
        mask_file = "ORCA_aligned_resampled_toPET_CT.nii.gz" if ct_file.endswith("0000.nii.gz") else "ORCA_aligned_mask.nii.gz"
        mask_CT_edges(ct_file, os.path.join(mask_dir, mask_file),edge_trim=50,top_trim=10)
    
    # Clip CT values lower than -1024 to -1024
    for ct_file in glob.glob(os.path.join(orca_working_dir, constants.ALIGNED_CT_FOLDER + "/*.nii*")):   
        ct_image = nibabel.load(ct_file)
        ct_image_data = ct_image.get_fdata()
        ct_image_data[ct_image_data < -1024] = -1024
        nibabel.save(nibabel.Nifti1Image(ct_image_data, ct_image.affine, ct_image.header), ct_file)


class ImageRegistration:
    def __init__(self, fixed_img: str, multi_resolution_iterations: str, num_workers: int, fixed_mask: str = None):
        self.fixed_img = fixed_img
        self.fixed_mask = fixed_mask
        self.multi_resolution_iterations = multi_resolution_iterations
        self.num_workers = num_workers
        self.moving_img = None
        self.transform_files = None

    def set_moving_image(self, moving_img: str, update_transforms: bool = True):
        self.moving_img = moving_img
        if update_transforms:
            out_dir = pathlib.Path(self.moving_img).parent
            moving_img_filename = pathlib.Path(self.moving_img).name
            self.transform_files = {
                'rigid': os.path.join(out_dir, f"{moving_img_filename}_rigid.mat"),
                'affine': os.path.join(out_dir, f"{moving_img_filename}_affine.mat"),
                'warp': os.path.join(out_dir, f"{moving_img_filename}_warp.nii.gz"),
                'inverse_warp': os.path.join(out_dir, f"{moving_img_filename}_inverse_warp.nii.gz")
            }

    def create_and_set_masked_moving_image(self, mask: str) -> float:
        """ Creates a masked moving image using the given mask
        :param mask: The mask to use
        """
        mask = nibabel.load(mask).get_fdata().astype(bool)
        print("self.moving image is ", self.moving_img, "\n") if DEBUG_MODE else None
        moving_image = nibabel.load(self.moving_img).get_fdata()
        # Set the background to the median value outside the mask
        median = np.median(moving_image[~mask])
        print("median used for masking is ", median, "\n") if DEBUG_MODE else None
        moving_image[~mask] = median
        nibabel.save(nibabel.Nifti1Image(moving_image, nibabel.load(self.moving_img).affine,nibabel.load(self.moving_img).header), \
                                    os.path.join(os.path.dirname(self.moving_img), 'masked_' + os.path.basename(self.moving_img)))
        self.masked_moving_image = os.path.join(os.path.dirname(self.moving_img), 'masked_' + os.path.basename(self.moving_img))
        print("self.masked_moving_image is ", self.masked_moving_image, "\n") if DEBUG_MODE else None
        return median 

    def rigid(self) -> str:
        mask_cmd = f"-gm {re.escape(self.fixed_mask)}" if self.fixed_mask else ""
        cmd_to_run = f"{GREEDY_PATH} -d 3 -a -i {re.escape(self.fixed_img)} {re.escape(self.masked_moving_image)} " \
                     f"{mask_cmd} -ia-image-centers -dof 6 -o {re.escape(self.transform_files['rigid'])} " \
                     f"-n {self.multi_resolution_iterations} -m NCC 2x2x2 -threads {self.num_workers}"
        print("Rigid: cmd_to_run is ", cmd_to_run, "\n") if DEBUG_MODE else None
        subprocess.run(cmd_to_run, shell=True, capture_output=True)
        logging.info(
            f"Rigid alignment: {pathlib.Path(self.masked_moving_image).name} -> {pathlib.Path(self.fixed_img).name} | Aligned image: "
            f"moco-{pathlib.Path(self.masked_moving_image).name} | Transform file: {pathlib.Path(self.transform_files['rigid']).name}")
        return self.transform_files['rigid']

    def affine(self) -> str:
        mask_cmd = f"-gm {re.escape(self.fixed_mask)}" if self.fixed_mask else ""
        cmd_to_run = f"{GREEDY_PATH} -d 3 -a -i {re.escape(self.fixed_img)} {re.escape(self.masked_moving_image)} " \
                     f"{mask_cmd} -ia-image-centers -dof 12 -o {re.escape(self.transform_files['affine'])} " \
                     f"-n {self.multi_resolution_iterations} -m NCC 2x2x2 -m NCC 2x2x2 -threads {self.num_workers}"
        subprocess.run(cmd_to_run, shell=True, capture_output=True)
        logging.info(
            f"Affine alignment: {pathlib.Path(self.masked_moving_image).name} -> {pathlib.Path(self.fixed_img).name} |"
            f" Aligned image: moco-{pathlib.Path(self.masked_moving_image).name} | Transform file: {pathlib.Path(self.transform_files['affine']).name}")
        return self.transform_files['affine']

    def deformable(self) -> tuple:
        self.rigid()
        mask_cmd = f"-gm {re.escape(self.fixed_mask)}" if self.fixed_mask else ""
        cmd_to_run = f"{GREEDY_PATH} -d 3 -m NCC 2x2x2 -i {re.escape(self.fixed_img)} {re.escape(self.masked_moving_image)} " \
                     f"{mask_cmd} -it {re.escape(self.transform_files['rigid'])} -o {re.escape(self.transform_files['warp'])} " \
                     f"-n {self.multi_resolution_iterations} " \
                     f"-threads {self.num_workers}"
        print("Deformable: cmd_to_run is ", cmd_to_run, "\n") if DEBUG_MODE else None
        subprocess.run(cmd_to_run, shell=True, capture_output=True)
        logging.info(
            f"Deformable alignment: {pathlib.Path(self.masked_moving_image).name} -> {pathlib.Path(self.fixed_img).name} | "
            f"Aligned image: moco-{pathlib.Path(self.masked_moving_image).name} | "
            f"Initial alignment:{pathlib.Path(self.transform_files['rigid']).name}"
            f" | warp file: {pathlib.Path(self.transform_files['warp']).name}")
        return self.transform_files['rigid'], self.transform_files['warp'], self.transform_files['inverse_warp']

    def registration(self, registration_type: str) -> None:
        if registration_type == 'rigid':
            self.rigid()
        elif registration_type == 'affine':
            self.affine()
        elif registration_type == 'deformable':
            self.deformable()
        else:
            sys.exit("Registration type not supported!")

    def resample(self, resampled_moving_img: str, registration_type: str, median=0.0, 
                 segmentation="", resampled_seg="") -> None:
        if registration_type == 'rigid':
            cmd_to_run = self._build_cmd(resampled_moving_img, segmentation, resampled_seg,
                                         self.transform_files['rigid'], median=median)
        elif registration_type == 'affine':
            cmd_to_run = self._build_cmd(resampled_moving_img, segmentation, resampled_seg,
                                         self.transform_files['affine'], median=median)
        elif registration_type == 'deformable':
            cmd_to_run = self._build_cmd(resampled_moving_img, segmentation, resampled_seg,
                                         self.transform_files['warp'], self.transform_files['rigid'], 
                                         median=median)
        print("Resample: cmd_to_run is ", cmd_to_run, "\n") if DEBUG_MODE else None
        subprocess.run(cmd_to_run, shell=True, capture_output=True)

    def _build_cmd(self, resampled_moving_img: str, segmentation: str, resampled_seg: str,
                   *transform_files: str, median: float) -> str:
        cmd = f"{GREEDY_PATH} -d 3 -rf {re.escape(self.fixed_img)} -ri LINEAR -rm " \
              f"{re.escape(self.moving_img)} {re.escape(resampled_moving_img)}"
        if segmentation and resampled_seg:
            cmd += f" -ri LABEL 0.2vox -rm {re.escape(segmentation)} {re.escape(resampled_seg)}"
        for transform_file in transform_files:
            cmd += f" -r {re.escape(transform_file)} -bg {median}"
        return cmd

def mask_CT_edges(ct_image: str, mask_file: str, edge_trim=10, top_trim=10) -> None:
    ct_image_nib = nibabel.load(ct_image)
    ct_image_data = ct_image_nib.get_fdata()
    mask = nibabel.load(mask_file).get_fdata()
    mask=mask.astype(bool)
    median = np.median(ct_image_data[~mask])

    for i in range(1,edge_trim):
        if (np.max(mask[:i,:,:])+np.max(mask[-i:,:,:])+np.max(mask[:,:i,:])+np.max(mask[:,-i:,:]))==0:
            ct_image_data[:i,:,:]=median
            ct_image_data[-i:,:,:]=median

            ct_image_data[:,:i,:]=median
            ct_image_data[:,-i:,:]=median
        else:
            break

    for i in range(1,top_trim):
        ct_image_data[:,:,-i][~mask[:,:,-i]]=median
        ct_image_data[:,:,i-1][~mask[:,:,i]]=median
    nibabel.save(nibabel.Nifti1Image(ct_image_data, ct_image_nib.affine, ct_image_nib.header), ct_image)


# def mask_CT(ct_dir: str, mask_dir: str, mask_id: str) -> None:
#     """
#     Masks the CT images in the ct_dir using the mask with the given identity
#     :param ct_dir: The directory containing the CT images
#     :param mask_dir: The directory containing the masks
#     :param mask_id: The identity of the mask to use
#     """
#     ct_files = sorted(glob.glob(os.path.join(ct_dir, 'resampled*.nii')))
#     mask =  glob.glob(os.path.join(mask_dir, f'{mask_id}*.nii*'))[0]
#     with Progress() as progress:
#         task = progress.add_task("[cyan] Masking CT files...", total=len(ct_files))

#         for ct_file in ct_files:
#             ct_image = nibabel.load(ct_file)
#             mask_image = nibabel.load(mask)
#             masked_image = nibabel.Nifti1Image(ct_image.get_fdata() * mask_image.get_fdata(), ct_image.affine)
#             nibabel.save(masked_image, os.path.join(ct_dir, f'{mask_id}_masked_' + os.path.basename(ct_file)))
#             progress.update(task, advance=1)

def get_mask(mask_dir: str, mask_id: str) -> str:
    """
    Gets the mask with the given identity
    :param mask_dir: The directory containing the masks
    :param mask_id: The identity of the mask to use
    :return: The path to the mask
    """
    return glob.glob(os.path.join(mask_dir, f'{mask_id}*.nii*'))[0]

def align(puma_working_dir: str, ct_dir: str, pt_dir: str, mask_dir: str, n_threads: int) -> None:
    reference_sCT_image = glob.glob(os.path.join(ct_dir, 'synth*.nii*'))[0]
    print("Using refference image of sCT: ", reference_sCT_image, "\n") if DEBUG_MODE else None
    moving_CT_image = glob.glob(os.path.join(ct_dir, 'resampled*.nii*'))[0]
    print("Using moving image of CT: ", moving_CT_image, "\n") if DEBUG_MODE else None
    mask = get_mask(mask_dir, 'dil8')

    with Progress() as progress:
        task = progress.add_task("[cyan] Aligning CT image to sCT", total=2)

        aligner = ImageRegistration(fixed_img=reference_sCT_image,
                                    multi_resolution_iterations=constants.MULTI_RESOLUTION_SCHEME,
                                    fixed_mask=mask, num_workers=n_threads)
        aligner.set_moving_image(moving_CT_image)
        median = aligner.create_and_set_masked_moving_image(get_mask(mask_dir, 'dil2'))
        aligner.registration('deformable')
        progress.update(task, advance=1)
        aligner.resample(resampled_moving_img=os.path.join(puma_working_dir, constants.ALIGNED_PREFIX +
                                                            os.path.basename(moving_CT_image)),
                            median=median, registration_type='deformable')
        
        # resample the mask to the new co-registered CT image
        aligner.resample(resampled_moving_img=os.path.join(puma_working_dir, constants.ALIGNED_PREFIX +
                                                            os.path.basename(moving_CT_image)),
                                                            segmentation=get_mask(mask_dir, 'resampled_'),
                                                            resampled_seg=os.path.join(mask_dir, constants.ALIGNED_PREFIX +
                                                            'resampled_toPET_CT.nii.gz'), median=median,
                            registration_type='deformable')
        
        # Clean up the edges of the resampled image
        # resampled_moving_img=os.path.join(puma_working_dir, constants.ALIGNED_PREFIX + os.path.basename(moving_CT_image))
        # ct_image = nibabel.load(resampled_moving_img)
        # ct_image_data = ct_image.get_fdata()
        # mask_file = os.path.join(mask_dir, constants.ALIGNED_PREFIX +'resampled_toPET_CT.nii.gz')
        # mask = nibabel.load(mask_file).get_fdata()
        # mask=mask.astype(bool)
        # median = np.median(ct_image_data[~mask])

        # for i in range(1,10):
        #     if (np.max(mask[:i,:,:])+np.max(mask[-i:,:,:])+np.max(mask[:,:i,:])+np.max(mask[:,-i:,:]))==0:
        #         ct_image_data[:i,:,:]=median
        #         ct_image_data[-i:,:,:]=median

        #         ct_image_data[:,:i,:]=median
        #         ct_image_data[:,-i:,:]=median
        #     else:
        #         break

        # for i in range(1,10):
        #     ct_image_data[:,:,-i][~mask[:,:,-i]]=median
        #     ct_image_data[:,:,i][~mask[:,:,i]]=median
        # nibabel.save(nibabel.Nifti1Image(ct_image_data, ct_image.affine), resampled_moving_img)

        
        progress.update(task, advance=1)


        # clean up transforms to a new folder
        rigid_transform_files = sorted(glob.glob(os.path.join(ct_dir, '*_rigid.mat')))
        warp_files = sorted(glob.glob(os.path.join(ct_dir, '*warp.nii.gz')))
        transforms_dir = os.path.join(puma_working_dir, constants.TRANSFORMS_FOLDER)
        create_directory(transforms_dir)
        # move all the warp files and rigid transform files to the transforms folder without zipping
        for rigid_transform_file in rigid_transform_files:
            move_file(rigid_transform_file, transforms_dir)
        for warp_file in warp_files:
            move_file(warp_file, transforms_dir)

        # move the aligned files to a new folder called aligned_CT, this is stored in the puma_working_dir
        aligned_ct_dir = os.path.join(puma_working_dir, constants.ALIGNED_CT_FOLDER)
        create_directory(aligned_ct_dir)
        # get aligned ct files using glob by looking for keyword 'aligned'
        aligned_ct_files = sorted(glob.glob(os.path.join(puma_working_dir, constants.ALIGNED_PREFIX + '*CT*.nii*')))
        for aligned_ct_file in aligned_ct_files:
            move_file(aligned_ct_file, aligned_ct_dir)

def export_dicom(puma_working_dir: str, subj_dir: str) -> None:
    print("Exporting to dicom \n") if DEBUG_MODE else None
    CT_dirs = glob.glob(os.path.join(subj_dir, '*_CT'))
    if len(CT_dirs)==1 and os.path.isdir(CT_dirs[0]):
        save_dicom_from_nifti_image(ref_dir=os.path.join(os.path.dirname(os.path.abspath(subj_dir)),CT_dirs[0]),
                                    nifti_path=glob.glob(os.path.join(puma_working_dir, constants.ALIGNED_CT_FOLDER, 'ORCA_aligned_CT.nii*'))[0],
                                    output_dir=os.path.join(os.path.dirname(os.path.abspath(puma_working_dir)),"ORCA_CT_DICOM"),
                                    series_description="orca co-registered CT",
                                    vendor="sms",
                                    force_overwrite=False)


def run_moose(ct_dir: str, mask_dir: str):
    """
    Runs moosez on the CT image in the ct_dir (EXPECTS ONLY ONE CT IMAGE)
    :param ct_dir: The directory containing the CT images
    :param mask_dir: The directory containing the masks
    """
    moose(model_name=constants.MOOSE_MODEL,
          input_dir=ct_dir,
          output_dir=mask_dir,
          accelerator=resources.check_cuda())
    # get the mask file
    mask_file = glob.glob(os.path.join(mask_dir, '*.nii*'))[0]
    # make every non-zero voxel in the mask file as 1 and the background as 0
    mask = nibabel.load(mask_file).get_fdata()
    mask[mask > 0] = 1
    # write the mask file with the same name
    nibabel.save(nibabel.Nifti1Image(mask, nibabel.load(mask_file).affine, nibabel.load(mask_file).header), mask_file)

def process_mask(mask_dir: str) -> None:
    """
    Processes the mask files in the mask_dir
    :param mask_dir: The directory containing the mask files
    """
    # ORCA is only using the mask of the resampled CT, the reference space is that of PET
    mask_files = sorted(glob.glob(os.path.join(mask_dir, '*.nii*')))
    # Rename to remove the "clin_CT_body_segmentation_"
    for mask_file in mask_files:
        mask_file = pathlib.Path(mask_file)
        new_name = mask_file.name.replace("clin_CT_body_segmentation_", "")
        new_path = mask_file.with_name(new_name)
        mask_file.rename(new_path)
    # Get the files with the new file-names
    mask_files = sorted(glob.glob(os.path.join(mask_dir, '*.nii*')))

    with Progress() as progress:
        task = progress.add_task("[cyan] Processing mask files ", total=len(mask_files))

        # For each mask create two masks, with binary dilations of 2 and 2+8
        for mask_file in mask_files:
            mask = nibabel.load(mask_file).get_fdata()
            mask[mask > 0] = 1
            mask_dil2 = binary_dilation(mask, iterations=2).astype(dtype=np.int32)
            mask_dil8 = binary_dilation(mask_dil2, iterations=8).astype(dtype=np.int32)
            nibabel.save(nibabel.Nifti1Image(mask_dil2, nibabel.load(mask_file).affine, \
                                             nibabel.load(mask_file).header),\
                                             os.path.join(mask_dir, 'dil2_' + os.path.basename(mask_file)))
            nibabel.save(nibabel.Nifti1Image(mask_dil8, nibabel.load(mask_file).affine, \
                                             nibabel.load(mask_file).header), \
                                             os.path.join(mask_dir, 'dil8_' + os.path.basename(mask_file)))
            progress.update(task, advance=1)




import torch
import torch.nn as nn
from torch.nn import init
import functools
from torch.optim import lr_scheduler


lrelu = nn.LeakyReLU(0.2)


class Identity(nn.Module):
    def forward(self, x):
        return x


def get_norm_layer(norm_type='instance'):
    """Return a normalization layer
    Parameters:
        norm_type (str) -- the name of the normalization layer: batch | instance | none
    For BatchNorm, we use learnable affine parameters and track running statistics (mean/stddev).
    For InstanceNorm, we do not use learnable affine parameters. We do not track running statistics.
    """
    if norm_type == 'batch':
        norm_layer = functools.partial(nn.BatchNorm3d, affine=True, track_running_stats=True)
    elif norm_type == 'instance':
        norm_layer = functools.partial(nn.InstanceNorm3d, affine=False, track_running_stats=False)
    elif norm_type == 'none':
        def norm_layer(x): return Identity()
    else:
        raise NotImplementedError('normalization layer [%s] is not found' % norm_type)
    return norm_layer


def init_weights(net, init_type='normal', init_gain=0.02):
    """Initialize network weights.
    Parameters:
        net (network)   -- network to be initialized
        init_type (str) -- the name of an initialization method: normal | xavier | kaiming | orthogonal
        init_gain (float)    -- scaling factor for normal, xavier and orthogonal.
    We use 'normal' in the original pix2pix and CycleGAN paper. But xavier and kaiming might
    work better for some applications. Feel free to try yourself.
    """
    def init_func(m):  # define the initialization function
        classname = m.__class__.__name__
        if hasattr(m, 'weight') and (classname.find('Conv') != -1 or classname.find('Linear') != -1):
            if init_type == 'normal':
                init.normal_(m.weight.data, 0.0, init_gain)
            elif init_type == 'xavier':
                init.xavier_normal_(m.weight.data, gain=init_gain)
            elif init_type == 'kaiming':
                init.kaiming_normal_(m.weight.data, a=0, mode='fan_in')
            elif init_type == 'orthogonal':
                init.orthogonal_(m.weight.data, gain=init_gain)
            else:
                raise NotImplementedError('initialization method [%s] is not implemented' % init_type)
            if hasattr(m, 'bias') and m.bias is not None:
                init.constant_(m.bias.data, 0.0)
        elif classname.find('BatchNorm3d') != -1:  # BatchNorm Layer's weight is not a matrix; only normal distribution applies.
            init.normal_(m.weight.data, 1.0, init_gain)
            init.constant_(m.bias.data, 0.0)

    # print('initialize network with %s' % init_type)
    net.apply(init_func)  # apply the initialization function <init_func>



# -----  ResnetGenerator -----------

class ResnetGenerator(nn.Module):
    """Resnet-based generator that consists of Resnet blocks between a few downsampling/upsampling operations.
    We adapt Torch code and idea from Justin Johnson's neural style transfer project(https://github.com/jcjohnson/fast-neural-style)
    """

    def __init__(self, input_nc, output_nc, ngf=64, norm_layer=nn.BatchNorm3d, use_dropout=False, n_blocks=3, padding_type='replicate'):
        """Construct a Resnet-based generator
        Parameters:
            input_nc (int)      -- the number of channels in input images
            output_nc (int)     -- the number of channels in output images
            ngf (int)           -- the number of filters in the last conv layer
            norm_layer          -- normalization layer
            use_dropout (bool)  -- if use dropout layers
            n_blocks (int)      -- the number of ResNet blocks
            padding_type (str)  -- the name of padding layer in conv layers: reflect | replicate | zero
        """
        assert(n_blocks >= 0)
        super(ResnetGenerator, self).__init__()
        if type(norm_layer) == functools.partial:
            use_bias = norm_layer.func == nn.InstanceNorm3d
        else:
            use_bias = norm_layer == nn.InstanceNorm3d

        # model = [nn.ReflectionPad2d(3),
        #          nn.Conv3d(input_nc, ngf, kernel_size=7, padding=0, bias=use_bias),
        #          norm_layer(ngf),
        #          nn.ReLU(True)]

        model = [nn.ReplicationPad3d(3),
                 nn.Conv3d(input_nc, ngf, kernel_size=7, padding=0, bias=use_bias),
                 norm_layer(ngf),
                 nn.ReLU(True)]

        n_downsampling = 2
        for i in range(n_downsampling):  # add downsampling layers
            mult = 2 ** i
            model += [nn.Conv3d(ngf * mult, ngf * mult * 2, kernel_size=3, stride=2, padding=1, bias=use_bias),
                      norm_layer(ngf * mult * 2),
                      nn.ReLU(True)]

        mult = 2 ** n_downsampling
        for i in range(n_blocks):       # add ResNet blocks

            model += [ResnetBlock(ngf * mult, padding_type=padding_type, norm_layer=norm_layer, use_dropout=use_dropout, use_bias=use_bias)]

        for i in range(n_downsampling):  # add upsampling layers
            mult = 2 ** (n_downsampling - i)
            model += [nn.ConvTranspose3d(ngf * mult, int(ngf * mult / 2),
                                         kernel_size=3, stride=2,
                                         padding=1, output_padding=1,
                                         bias=use_bias),
                      norm_layer(int(ngf * mult / 2)),
                      nn.ReLU(True)]
        # model += [nn.ReflectionPad2d(3)]
        model += [nn.ReplicationPad3d(3)]
        model += [nn.Conv3d(ngf, output_nc, kernel_size=7, padding=0)]
        model += [nn.Tanh()]


        self.model = nn.Sequential(*model)

    def forward(self, input):
        """Standard forward"""
        return self.model(input)


class ResnetBlock(nn.Module):
    """Define a Resnet block"""

    def __init__(self, dim, padding_type, norm_layer, use_dropout, use_bias):
        """Initialize the Resnet block
        A resnet block is a conv block with skip connections
        We construct a conv block with build_conv_block function,
        and implement skip connections in <forward> function.
        Original Resnet paper: https://arxiv.org/pdf/1512.03385.pdf
        """
        super(ResnetBlock, self).__init__()
        self.conv_block = self.build_conv_block(dim, padding_type, norm_layer, use_dropout, use_bias)

    def build_conv_block(self, dim, padding_type, norm_layer, use_dropout, use_bias):
        """Construct a convolutional block.
        Parameters:
            dim (int)           -- the number of channels in the conv layer.
            padding_type (str)  -- the name of padding layer: reflect | replicate | zero
            norm_layer          -- normalization layer
            use_dropout (bool)  -- if use dropout layers.
            use_bias (bool)     -- if the conv layer uses bias or not
        Returns a conv block (with a conv layer, a normalization layer, and a non-linearity layer (ReLU))
        """
        conv_block = []
        p = 0
        if padding_type == 'reflect':
            conv_block += [nn.ReflectionPad2d(1)]
        elif padding_type == 'replicate':
            conv_block += [nn.ReplicationPad3d(1)]
        elif padding_type == 'zero':
            p = 1
        else:
            raise NotImplementedError('padding [%s] is not implemented' % padding_type)

        conv_block += [nn.Conv3d(dim, dim, kernel_size=3, padding=p, bias=use_bias), norm_layer(dim), nn.ReLU(True)]
        if use_dropout:
            conv_block += [nn.Dropout(0.5)]

        p = 0
        if padding_type == 'reflect':
            conv_block += [nn.ReflectionPad2d(1)]
        elif padding_type == 'replicate':
            conv_block += [nn.ReplicationPad3d(1)]
        elif padding_type == 'zero':
            p = 1
        else:
            raise NotImplementedError('padding [%s] is not implemented' % padding_type)
        conv_block += [nn.Conv3d(dim, dim, kernel_size=3, padding=p, bias=use_bias), norm_layer(dim)]

        return nn.Sequential(*conv_block)

    def forward(self, x):
        """Forward function (with skip connections)"""
        out = x + self.conv_block(x)  # add skip connections
        return out

# -------------------------------

class UnetGenerator(nn.Module):
    """Create a Unet-based generator"""

    def __init__(self, input_nc=1, output_nc=1, num_downs=6, ngf=64, norm_layer=nn.BatchNorm3d, use_dropout=True):
        super(UnetGenerator, self).__init__()

        # currently support only input_nc == output_nc
        assert (input_nc == output_nc)

        # construct unet structure
        unet_block = UnetSkipConnectionBlock(ngf * 8, ngf * 8, input_nc=None, submodule=None, norm_layer=norm_layer, innermost=True)  # add the innermost layer
        for i in range(num_downs - 5):          # add intermediate layers with ngf * 8 filters
            unet_block = UnetSkipConnectionBlock(ngf * 8, ngf * 8, input_nc=None, submodule=unet_block, norm_layer=norm_layer, use_dropout=use_dropout)
        # gradually reduce the number of filters from ngf * 8 to ngf
        unet_block = UnetSkipConnectionBlock(ngf * 4, ngf * 8, input_nc=None, submodule=unet_block, norm_layer=norm_layer)
        unet_block = UnetSkipConnectionBlock(ngf * 2, ngf * 4, input_nc=None, submodule=unet_block, norm_layer=norm_layer)
        unet_block = UnetSkipConnectionBlock(ngf, ngf * 2, input_nc=None, submodule=unet_block, norm_layer=norm_layer)
        self.model = UnetSkipConnectionBlock(output_nc, ngf, input_nc=input_nc, submodule=unet_block, outermost=True, norm_layer=norm_layer)  # add the outermost layer

    def forward(self, input):
        """Standard forward"""
        return self.model(input)


class UnetSkipConnectionBlock(nn.Module):
    """Defines the Unet submodule with skip connection.
        X -------------------identity----------------------
        |-- downsampling -- |submodule| -- upsampling --|
    """

    def __init__(self, outer_nc, inner_nc, input_nc=None,
                 submodule=None, outermost=False, innermost=False, norm_layer=nn.BatchNorm3d, use_dropout=False):
        """Construct a Unet submodule with skip connections.
        Parameters:
            outer_nc (int) -- the number of filters in the outer conv layer
            inner_nc (int) -- the number of filters in the inner conv layer
            input_nc (int) -- the number of channels in input images/features
            submodule (UnetSkipConnectionBlock) -- previously defined submodules
            outermost (bool)    -- if this module is the outermost module
            innermost (bool)    -- if this module is the innermost module
            norm_layer          -- normalization layer
            use_dropout (bool)  -- if use dropout layers.
        """
        super(UnetSkipConnectionBlock, self).__init__()
        self.outermost = outermost
        if type(norm_layer) == functools.partial:
            use_bias = norm_layer.func == nn.BatchNorm3d
        else:
            use_bias = norm_layer == nn.BatchNorm3d
        if input_nc is None:
            input_nc = outer_nc
        downconv = nn.Conv3d(input_nc, inner_nc, kernel_size=4, stride=2, padding=1, bias=use_bias)
        downrelu = nn.LeakyReLU(0.2, True)
        downnorm = norm_layer(inner_nc)
        uprelu = nn.ReLU(True)
        upnorm = norm_layer(outer_nc)

        if outermost:
            upconv = nn.ConvTranspose3d(inner_nc * 2, outer_nc, kernel_size=4, stride=2, padding=1)
            down = [downconv]
            up = [uprelu, upconv, nn.Tanh()]
            model = down + [submodule] + up
        elif innermost:
            upconv = nn.ConvTranspose3d(inner_nc, outer_nc, kernel_size=4, stride=2, padding=1, bias=use_bias)
            down = [downrelu, downconv]
            up = [uprelu, upconv, upnorm]
            model = down + up
        else:
            upconv = nn.ConvTranspose3d(inner_nc * 2, outer_nc, kernel_size=4, stride=2, padding=1, bias=use_bias)
            down = [downrelu, downconv, downnorm]
            up = [uprelu, upconv, upnorm]

            if use_dropout:
                model = down + [submodule] + up + [nn.Dropout(0.5)]
            else:
                model = down + [submodule] + up

        self.model = nn.Sequential(*model)

    def forward(self, x):
        if self.outermost:
            return self.model(x)
        else:   # add skip connections
            return torch.cat([x, self.model(x)], 1)

class conv_block(nn.Module):
    def __init__(self, in_c, out_c):
        super().__init__()
        self.conv1 = nn.Conv3d(in_c, out_c, kernel_size=3, padding=1, stride=1)
        self.bn1 = nn.BatchNorm3d(out_c)         
        self.conv2 = nn.Conv3d(out_c, out_c, kernel_size=3, padding=1,stride=1)
        self.bn2 = nn.BatchNorm3d(out_c)         
        self.relu = nn.PReLU() 
        # identity convolution
        self.idconv =  nn.Conv3d(in_c, out_c, kernel_size=1)   
    
    def forward(self, inputs):
        identity = self.idconv(inputs)
        x = self.conv1(inputs)
        x = self.bn1(x)
        x = self.relu(x)        
        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu(x)
        return x + identity
        
class encoder_block(nn.Module):
    def __init__(self, in_c, out_c):
        super().__init__()
        self.conv = conv_block(in_c, out_c)
        self.pool = nn.AvgPool3d((2, 2, 2))
    def forward(self, inputs):
        x = self.conv(inputs)
        p = self.pool(x)
        return x, p

class decoder_block(nn.Module):
    def __init__(self, in_c, out_c):
        super().__init__()
        self.up = nn.ConvTranspose3d(in_c, out_c, kernel_size=2, stride=2)
        self.conv = conv_block(out_c+out_c, out_c)     
    def forward(self, inputs, skip):
        x = self.up(inputs)
        x = torch.cat([x, skip], axis=1)
        x = self.conv(x)
        return x

class build_resunet(nn.Module):
    def __init__(self,ngf):
        super().__init__()
        """ Encoder """
        self.e1 = encoder_block(1, ngf)
        self.e2 = encoder_block(ngf, ngf*2)
        self.e3 = encoder_block(ngf*2, ngf*4)
        self.e4 = encoder_block(ngf*4, ngf*8)         
        """ Bottleneck """
        self.b = conv_block(ngf*8, ngf*16)         
        """ Decoder """
        self.d1 = decoder_block(ngf*16, ngf*8)
        self.d2 = decoder_block(ngf*8, ngf*4)
        self.d3 = decoder_block(ngf*4, ngf*2)
        self.d4 = decoder_block(ngf*2, ngf)         
        """ Output """
        self.outputs = nn.Conv3d(ngf, 1, kernel_size=1, padding=0)
        self.final_act = nn.Tanh()     
    def forward(self, inputs):
        """ Encoder """
        s1, p1 = self.e1(inputs)
        s2, p2 = self.e2(p1)
        s3, p3 = self.e3(p2)
        s4, p4 = self.e4(p3) 
        """ Bottleneck """
        b = self.b(p4)  
        """ Decoder """
        d1 = self.d1(b, s4)
        d2 = self.d2(d1, s3)
        d3 = self.d3(d2, s2)
        d4 = self.d4(d3, s1)         
        """ Output """
        outputs = self.outputs(d4)

        return self.final_act(outputs)

def build_netG(opt):
    if opt.netG == 'resnet':
        generator = ResnetGenerator(opt.img_channel, opt.img_channel, opt.ngf, use_dropout=True, n_blocks=9)
    elif opt.netG == 'Unet':
        generator = UnetGenerator(ngf=opt.ngf, use_dropout=True, norm_layer=nn.BatchNorm3d)
    else:
        raise NotImplementedError

    init_weights(generator, init_type='normal')

    return generator


# if __name__ == '__main__':
#     import time
#     import torch
#     from torch.autograd import Variables

#     from init import Options
#     opt = Options().parse()

#     torch.cuda.set_device(0)
#     generator = build_netG(opt)

#     net = generator.cuda().eval()

#     data = Variable(torch.randn(int(opt.batch_size), int(opt.img_channel), int(opt.patch_size[0]), int(opt.patch_size[1]), int(opt.patch_size[2]))).cuda()

#     out = net(data)

#     print("out size: {}".format(out.size()))












#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# Author: Lalith Kumar Shiyam Sundar | Sebastian Gutschmayer
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 04.07.2023
# Version: 0.1.0
#
# Description:
# This module downloads the necessary binaries and models for the orca.
#
# Usage:
# The functions in this module can be imported and used in other modules within the orca to download the necessary
# binaries and models for the orca.
#
# ----------------------------------------------------------------------------------------------------------------------

import logging
import os
import zipfile

import requests
from orcaz import file_utilities
from orcaz import constants
import falconz.resources as falcon_resources

from rich.console import Console
from rich.progress import Progress, BarColumn, TextColumn, TimeRemainingColumn, FileSizeColumn, TransferSpeedColumn
import time
import emoji

def download_binaries():
    # ----------------------------------
    # DOWNLOADING THE BINARIES
    # ----------------------------------

    print('')
    print(f'{constants.ANSI_BLUE} {emoji.emojize(":globe_with_meridians:")} BINARIES DOWNLOAD:{constants.ANSI_RESET}')

    print('')
    binary_path = constants.BINARY_PATH
    file_utilities.create_directory(binary_path)
    system_os, system_arch = file_utilities.get_system()
    print(f'{constants.ANSI_ORANGE} Detected system: {system_os} | Detected architecture: {system_arch}'
          f'{constants.ANSI_RESET}')
    download(item_name=f'falcon-{system_os}-{system_arch}', item_path=binary_path,
                      item_dict=falcon_resources.FALCON_BINARIES)
    file_utilities.set_permissions(constants.GREEDY_PATH, system_os)
    file_utilities.set_permissions(constants.DCM2NIIX_PATH, system_os)
    file_utilities.set_permissions(constants.C3D_PATH, system_os)


def download(item_name, item_path, item_dict):
    """
    Downloads the item (model or binary) for the current system.
    :param item_name: The name of the item to download.
    :param item_path: The path to store the item.
    :param item_dict: The dictionary containing item info.
    """
    item_info = item_dict[item_name]
    url = item_info["url"]
    filename = os.path.join(item_path, item_info["filename"])
    directory = os.path.join(item_path, item_info["directory"])

    if not os.path.exists(directory):
        logging.info(f" Downloading {directory}")

        # show progress using rich
        response = requests.get(url, stream=True)
        total_size = int(response.headers.get("Content-Length", 0))
        chunk_size = 1024 * 10

        console = Console()
        progress = Progress(
            TextColumn("[bold blue]{task.description}"),
            BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>3.0f}%",
            "•",
            FileSizeColumn(),
            TransferSpeedColumn(),
            TimeRemainingColumn(),
            console=console,
            expand=True
        )

        Download_Message = f"ORCA specific executables" if "falcon" in item_name else f"sCT models for {item_name}"

        with progress:  
            task = progress.add_task(f"Downloading {Download_Message}", total=total_size)
            for chunk in response.iter_content(chunk_size=chunk_size):
                open(filename, "ab").write(chunk)
                progress.update(task, advance=chunk_size)

        # Unzip the item
        progress = Progress(  # Create new instance for extraction task
            TextColumn("[bold blue]{task.description}"),
            BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>3.0f}%",
            "•",
            FileSizeColumn(),
            TransferSpeedColumn(),
            TimeRemainingColumn(),
            console=console,
            expand=True
        )

        with progress:
            with zipfile.ZipFile(filename, 'r') as zip_ref:
                total_size = sum((file.file_size for file in zip_ref.infolist()))
                task = progress.add_task(f"[white] Extracting {Download_Message}",
                                         total=total_size)
                # Get the parent directory of 'directory'
                parent_directory = os.path.dirname(directory)
                for file in zip_ref.infolist():
                    zip_ref.extract(file, parent_directory)
                    extracted_size = file.file_size
                    progress.update(task, advance=extracted_size)

        logging.info(f" {os.path.basename(directory)} extracted.")

        # Delete the zip file
        os.remove(filename)
        print(f"{constants.ANSI_GREEN} Download complete. {constants.ANSI_RESET}")
        logging.info(f" Download complete.")
    else:
        Message = f"ORCA specific executables" if "falcon" in item_name else f"sCT models for {item_name}"
        print(f"{constants.ANSI_GREEN} A local instance of the {Message} has been detected. "
              f"{constants.ANSI_RESET}")
        logging.info(f" A local instance of the {Message} has been detected.")

    return os.path.join(item_path, item_name)



from setuptools import setup

setup(
    name='orcaz',
    version='0.1.5',
    packages=['orcaz'],
    url='',
    license='GNU General Public License v3.0',
    author='Zacharias Chalampalakis, PhD, Lalith Kumar Shiyam Sundar, PhD',
    author_email='zacharias.chalampalakis@meduniwien.ac.at, lalith.shiyamsundar@meduniwien.ac.at',
    description='ORCA (Optimized Registration through Conditional Adversarial networks) ',
    long_description_content_type="text/markdown",
    long_description=open('README.md').read(),
    install_requires=['pyfiglet~=0.8.post1',
                      'setuptools~=65.5.1',
                      'nibabel',
                      'tqdm',
                      'torch>=0.4.1',
                      'torchvision>=0.2.1',
                      'matplotlib',
                      'tensorboard',
                      'scipy',
                      'SimpleITK',
                      'scikit-learn',
                      'emoji',
                      'pydicom',
                      'dicom2nifti',
                      'moosez>3.0.0',
                      'falconz',
                      'nifti2dicom',
                      'rich',
                      'pathlib'],
    entry_points={
        'console_scripts': [
            'orcaz = orcaz.orcaz:main',
        ],
    },
)




#!/usr/bin/env python2
# -*- coding: utf-8 -*-
from tqdm import tqdm
import datetime
import argparse
import matplotlib.pyplot as plt
import math
import scipy
import torch
from torch.autograd import Variable
import sys
import os
import torch
from utils.utils import *
import torch.optim as optim
import torch.nn as nn
from generators import build_netG
from discriminators import build_netD
from utils.NiftiDataset import *
import utils.NiftiDataset as NiftiDataset
from torch.utils.data import DataLoader
from collections import OrderedDict
from glob import glob
from pathlib import Path


parser = argparse.ArgumentParser()
parser.add_argument('--multi_gpu', default=True, help='Multi or single Gpu')
parser.add_argument('--gpu_ids', default='0', help='Select the GPU')
parser.add_argument("--image", type=str, default='./Data_folder/VPat_002_STD_PET_Raw_Data-14-PSFTOF_000_000_it_4.nii')
parser.add_argument("--label", type=str, default=None)
parser.add_argument("--result", type=str, default='./Data_folder/VPat_002_STD_PET_Raw_Data-14-PSFTOF_000_000_it_4_Result.nii', help='path to the .nii result to save')
parser.add_argument("--weights", type=str, default='./checkpoints/g_epoch_500.pth', help='generator weights to load')
parser.add_argument("--resample", default=False, help='Decide or not to resample the images to a new resolution')
parser.add_argument("--new_resolution", type=float, default=(0.625, 0.625, 1), help='New resolution')
parser.add_argument("--patch_size", type=int, nargs=3, default=[128, 128, 128], help="Input dimension for the generator")
parser.add_argument("--batch_size", type=int, nargs=1, default=1, help="Batch size to feed the network (currently supports 1)")
parser.add_argument("--stride_inplane", type=int, nargs=1, default=16, help="Stride size in 2D plane")
parser.add_argument("--stride_layer", type=int, nargs=1, default=16, help="Stride size in z direction")
args = parser.parse_args()

def from_numpy_to_itk(image_np,image_itk):
    image_np = np.transpose(image_np, (2, 1, 0))
    image = sitk.GetImageFromArray(image_np)
    image.SetOrigin(image_itk.GetOrigin())
    image.SetDirection(image_itk.GetDirection())
    image.SetSpacing(image_itk.GetSpacing())
    return image


def prepare_batch(image, ijk_patch_indices):
    image_batches = []
    for batch in ijk_patch_indices:
        image_batch = []
        for patch in batch:
            image_patch = image[patch[0]:patch[1], patch[2]:patch[3], patch[4]:patch[5]]
            image_batch.append(image_patch)

        image_batch = np.asarray(image_batch)
        # image_batch = image_batch[:, :, :, :, np.newaxis]
        image_batches.append(image_batch)

    return image_batches


# inference single image
def inference(write_image, model, image_path, label_path, result_path, resample, resolution, patch_size_x, patch_size_y, patch_size_z, stride_inplane, stride_layer, batch_size=1, segmentation=True, Logger=True):


    # create transformations to image and labels
    transforms1 = [
        NiftiDataset.Resample(resolution, resample)
    ]

    transforms2 = [
        NiftiDataset.Padding((patch_size_x, patch_size_y, patch_size_z))
    ]

    # read image file
    reader = sitk.ImageFileReader()
    reader.SetFileName(image_path)
    image = reader.Execute()

    # normalize the image
    image = Normalization(image)

    castImageFilter = sitk.CastImageFilter()
    castImageFilter.SetOutputPixelType(sitk.sitkFloat32)
    image = castImageFilter.Execute(image)

    # create empty label in pair with transformed image
    label_tfm = sitk.Image(image.GetSize(), sitk.sitkFloat32)
    label_tfm.SetOrigin(image.GetOrigin())
    label_tfm.SetDirection(image.GetDirection())
    label_tfm.SetSpacing(image.GetSpacing())

    sample = {'image': image, 'label': label_tfm}

    for transform in transforms1:
        sample = transform(sample)

    # keeping track on how much padding will be performed before the inference
    image_array = sitk.GetArrayFromImage(sample['image'])
    pad_x = patch_size_x - (patch_size_x - image_array.shape[2])
    pad_y = patch_size_x - (patch_size_y - image_array.shape[1])
    pad_z = patch_size_z - (patch_size_z - image_array.shape[0])

    image_pre_pad = sample['image']

    for transform in transforms2:
        sample = transform(sample)

    image_tfm, label_tfm = sample['image'], sample['label']

    # convert image to numpy array
    image_np = sitk.GetArrayFromImage(image_tfm)
    label_np = sitk.GetArrayFromImage(label_tfm)

    label_np = np.asarray(label_np, np.float32)

    # unify numpy and sitk orientation
    image_np = np.transpose(image_np, (2, 1, 0))
    label_np = np.transpose(label_np, (2, 1, 0))

    if segmentation is True:
        label_np = np.around(label_np)

    # ----------------- Padding the image if the z dimension still is not even ----------------------

    if (image_np.shape[2] % 2) == 0:
        Padding = False
    else:
        image_np = np.pad(image_np, ((0,0), (0,0), (0, 1)), 'edge')
        label_np = np.pad(label_np, ((0, 0), (0, 0), (0, 1)), 'edge')
        Padding = True

    # ------------------------------------------------------------------------------------------------

    # a weighting matrix will be used for averaging the overlapped region
    weight_np = np.zeros(label_np.shape)

    # prepare image batch indices
    inum = int(math.ceil((image_np.shape[0] - patch_size_x) / float(stride_inplane))) + 1
    jnum = int(math.ceil((image_np.shape[1] - patch_size_y) / float(stride_inplane))) + 1
    knum = int(math.ceil((image_np.shape[2] - patch_size_z) / float(stride_layer))) + 1

    patch_total = 0
    ijk_patch_indices = []
    ijk_patch_indicies_tmp = []

    for i in range(inum):
        for j in range(jnum):
            for k in range(knum):
                if patch_total % batch_size == 0:
                    ijk_patch_indicies_tmp = []

                istart = i * stride_inplane
                if istart + patch_size_x > image_np.shape[0]:  # for last patch
                    istart = image_np.shape[0] - patch_size_x
                iend = istart + patch_size_x

                jstart = j * stride_inplane
                if jstart + patch_size_y > image_np.shape[1]:  # for last patch
                    jstart = image_np.shape[1] - patch_size_y
                jend = jstart + patch_size_y

                kstart = k * stride_layer
                if kstart + patch_size_z > image_np.shape[2]:  # for last patch
                    kstart = image_np.shape[2] - patch_size_z
                kend = kstart + patch_size_z

                ijk_patch_indicies_tmp.append([istart, iend, jstart, jend, kstart, kend])

                if patch_total % batch_size == 0:
                    ijk_patch_indices.append(ijk_patch_indicies_tmp)

                patch_total += 1

    batches = prepare_batch(image_np, ijk_patch_indices)

    if Logger is True:

        for i in tqdm(range(len(batches))):
            batch = batches[i]

            batch = (batch - 127.5) / 127.5

            batch = torch.from_numpy(batch[np.newaxis, :, :, :])
            batch = Variable(batch.cuda())

            pred = model(batch)
            pred = pred.squeeze().data.cpu().numpy()

            pred = (pred * 127.5) + 127.5

            istart = ijk_patch_indices[i][0][0]
            iend = ijk_patch_indices[i][0][1]
            jstart = ijk_patch_indices[i][0][2]
            jend = ijk_patch_indices[i][0][3]
            kstart = ijk_patch_indices[i][0][4]
            kend = ijk_patch_indices[i][0][5]
            label_np[istart:iend, jstart:jend, kstart:kend] += pred[:, :, :]
            weight_np[istart:iend, jstart:jend, kstart:kend] += 1.0
    else:

        for i in tqdm(range(len(batches))):
            batch = batches[i]

            batch = (batch - 127.5) / 127.5

            batch = torch.from_numpy(batch[np.newaxis, :, :, :])
            batch = Variable(batch.cuda())

            pred = model(batch)
            pred = pred.squeeze().data.cpu().numpy()

            pred = (pred * 127.5) + 127.5

            istart = ijk_patch_indices[i][0][0]
            iend = ijk_patch_indices[i][0][1]
            jstart = ijk_patch_indices[i][0][2]
            jend = ijk_patch_indices[i][0][3]
            kstart = ijk_patch_indices[i][0][4]
            kend = ijk_patch_indices[i][0][5]
            label_np[istart:iend, jstart:jend, kstart:kend] += pred[:, :, :]
            weight_np[istart:iend, jstart:jend, kstart:kend] += 1.0

        print("{}: Evaluation complete".format(datetime.datetime.now()))

    # eliminate overlapping region using the weighted value
    label_np = (np.float32(label_np) / np.float32(weight_np) + 0.01)

    if segmentation is True:
        label_np = abs(np.around(label_np))

    # removed the 1 pad on z
    if Padding is True:
        label_np = label_np[:, :, 0:(label_np.shape[2]-1)]

    # removed all the padding
    label_np = label_np[:pad_x, :pad_y, :pad_z]

    # convert back to sitk space
    label = from_numpy_to_itk(label_np, image_pre_pad)
    # ---------------------------------------------------------------------------------------------

    # save label
    writer = sitk.ImageFileWriter()

    if resample is True:

        print("{}: Resampling label back to original image space...".format(datetime.datetime.now()))
        # label = resample_sitk_image(label, spacing=image.GetSpacing(), interpolator='bspline')   # keep this commented
        if segmentation is True:
            label = resize(label, (sitk.GetArrayFromImage(image)).shape[::-1], sitk.sitkLinear)
            label_array = np.around(sitk.GetArrayFromImage(label))
            label = sitk.GetImageFromArray(label_array)
            label.SetDirection(image.GetDirection())
            label.SetOrigin(image.GetOrigin())
            label.SetSpacing(image.GetSpacing())
        else:
            label = resize(label, (sitk.GetArrayFromImage(image)).shape[::-1], sitk.sitkBSpline)
            label.SetDirection(image.GetDirection())
            label.SetOrigin(image.GetOrigin())
            label.SetSpacing(image.GetSpacing())


    else:
        label = label

    if label_path is not None and segmentation is True:

        reader = sitk.ImageFileReader()
        reader.SetFileName(label_path)
        true_label = reader.Execute()

        true_label = sitk.GetArrayFromImage(true_label)
        predicted = sitk.GetArrayFromImage(label)

        dice = dice_coeff(predicted,true_label)


    writer.SetFileName(result_path)
    if write_image is True:
        writer.Execute(label)
        print("{}: Save evaluate label at {} success".format(datetime.datetime.now(), result_path))

    if label_path is not None and segmentation is True:
        print("Dice score:", dice)

        return label, dice

    else:
        dice = None
        return label, dice
    

def images_inference(net,model):
    id = Path(model).stem
    for image in glob("./for_eval/test/*/*PET*"):
        print(image)
        print(image.replace("PET","sCT_"+id))
        result, dice = inference(True, net, image, None, (image.replace("PET","sCT_"+id)), args.resample, args.new_resolution,
                args.patch_size[0],args.patch_size[1],args.patch_size[2], args.stride_inplane, args.stride_layer, segmentation=False, Logger=False)



if __name__ == "__main__":

    from init import Options
    opt = Options().parse()

    for model in glob("./for_eval/*.pth"):
        print(model)
        if model.find("MONAI_270_Norm01")>0:
            device = torch.device("cuda:0")
            net = UNet(
                spatial_dims=3,
                in_channels=1,
                out_channels=1,
                channels=(64, 128, 256, 512, 1024),
                strides=(2, 2, 2, 1),
                num_res_units=2,
                norm=Norm.BATCH,
            ).to(device)

            net.load_state_dict(new_state_dict(model))
            images_inference(net,model)
        
        # if model.find("UNET")>0:
        #     opt.netG = 'Unet'
        #     net = build_netG(opt).cuda()
        #     net.load_state_dict(new_state_dict(model))
        #     images_inference(net,model)           
        
        # if model.find("Res")>0:
        #     opt.netG = 'resnet'
        #     net = build_netG(opt).cuda()
        #     net.load_state_dict(new_state_dict(model))
        #     images_inference(net,model)  



    # result, dice = inference(True, net, args.image, None, args.result, args.resample, args.new_resolution,
    #                    args.patch_size[0],args.patch_size[1],args.patch_size[2], args.stride_inplane, args.stride_layer, segmentation=False, Logger=False)
    
    # from logger import *
    # val_list = create_list(opt.val_path)
    # plot_generated_batch(val_list=val_list, model=net, resample=opt.resample, resolution=opt.new_resolution,
    #                     patch_size_x=opt.patch_size[0], patch_size_y=opt.patch_size[1],
    #                     patch_size_z=opt.patch_size[2], stride_inplane=opt.stride_inplane,
    #                     stride_layer=opt.stride_layer, batch_size=1,
    #                     epoch=900)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------------------------------------------------
# Author: Lalith Kumar Shiyam Sundar | Sebastian Gutschmayer
# Institution: Medical University of Vienna
# Research Group: Quantitative Imaging and Medical Physics (QIMP) Team
# Date: 04.07.2023
# Version: 0.1.0
#
# Description:
# This module contains the urls and filenames of the binaries that are required for the orca.
#
# Usage:
# The variables in this module can be imported and used in other modules within the orca to download the necessary
# binaries for the orca.
#
# ----------------------------------------------------------------------------------------------------------------------
import torch
from orcaz import constants

# ORCA MODELS

ORCA_MODELS = {
    "nac_fluorodeoxyglucose": {
        "url": "https://enhance-pet.s3.eu-central-1.amazonaws.com/orca/fdg_nac_g_epoch_2000.zip",
        "filename": "fdg_nac_g_epoch_2000.zip",
        "model": "fdg_nac_g_epoch_2000.pth",
        "directory": "fdg_nac_g_epoch_2000",
        "voxel_spacing": [2.80374, 2.80374, 2.00086]
    },
    "nac_psma": {
        "url": "https://enhance-pet.s3.eu-central-1.amazonaws.com/orca/psma_nac_g_epoch_2000.zip",
        "filename": "psma_nac_g_epoch_2000.zip",
        "model": "psma_nac_g_epoch_2000.pth",
        "directory": "psma_nac_g_epoch_2000",
        "voxel_spacing": [3.3, 3.3, 2]
    },
    "nac_fluciclovine": {
        "url": "https://enhance-pet.s3.eu-central-1.amazonaws.com/orca/fluciclovine_nac_g_epoch_2000.zip",
        "filename": "fluciclovine_nac_g_epoch_2000.zip",
        "model": "fluciclovine_nac_g_epoch_2000.pth",
        "directory": "fluciclovine_nac_g_epoch_2000",
        "voxel_spacing": [2.34375, 2.34375, 2.344]
    },
    "nac_dotatate": {
        "url": "https://enhance-pet.s3.eu-central-1.amazonaws.com/orca/dotatate_nac_g_epoch_2000.zip",
        "filename": "dotatate_nac_g_epoch_2000.zip",
        "model": "dotatate_nac_g_epoch_2000.pth",
        "directory": "dotatate_nac_g_epoch_2000",
        "voxel_spacing": [2.34375, 2.34375, 2.344]
    },
    "nac_agnostic": {
        "url": "https://enhance-pet.s3.eu-central-1.amazonaws.com/orca/agnostic_nac_g_epoch_2000.zip",
        "filename": "agnostic_nac_g_epoch_2000.zip",
        "model": "agnostic_nac_g_epoch_2000.pth",
        "directory": "agnostic_nac_g_epoch_2000",
        "voxel_spacing": [2.80374, 2.80374, 2.00086]
    },
}


def check_cuda(print_flag=False) -> str:
    """
    This function checks if CUDA is available on the device and prints the device name and number of CUDA devices
    available on the device.

    Returns:
        str: The device to run predictions on, either "cpu" or "cuda".
    """
    if not torch.cuda.is_available():
        print(
            f"{constants.ANSI_ORANGE}CUDA not available on this device. Predictions will be run on CPU.{constants.ANSI_RESET}") if print_flag else None
        return "cpu"
    else:
        device_count = torch.cuda.device_count()
        print(
            f"{constants.ANSI_GREEN} CUDA is available on this device with {device_count} GPU(s). Predictions will be run on GPU.{constants.ANSI_RESET}") if print_flag else None
        return "cuda"
